import importlib, yaml, pkg_resources, os, warnings
import torch 
import bpyth as bpy
import pandas as pd
import pandasklar as pak

    
#############################################################################################################
###
### Helper for Ludwig
###
#############################################################################################################    


def gpu_info():
    # GPU und CUDA
    if torch.cuda.is_available():
        print('CUDA is available on ' + torch.cuda.get_device_name(torch.cuda.current_device()))
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        torch.cuda.empty_cache() # Cache leeren
        print('max_memory_allocated: {}'.format(bpy.human_readable_bytes(torch.cuda.max_memory_allocated(device=device)) )  )
        print('max_memory_reserved:  {}'.format(bpy.human_readable_bytes(torch.cuda.max_memory_reserved( device=device)) )  )  

    else:
        print('no CUDA!')    



def entwirre_test_stat(test_stats):
    result = []
    for config_no, test_stat in enumerate(test_stats):
        print(config_no)
        for output_feature_name, stat in test_stat.items():
            if output_feature_name == 'combined':
                continue
            for key, value in stat.items():
                if isinstance(value, (int, float)):
                    result += [[config_no, key, bpy.human_readable_number(value,3) ]]    
                    
    result = pak.dataframe(result)
    result.columns = ['config_no','name','value']
    return result




def prepare_train_log(train_log_raw, size='small'):

    result = pd.pivot_table( train_log_raw, 
                              index='name',
                              columns='config_no',
                              values='value', 
                              aggfunc='first')
    result = pak.drop_multiindex(result).reset_index() 
    
    # Namen korrigieren
    mask = result.name.str.startswith('average')
    result.loc[mask,'name'] = result[mask].name.str.replace('average_','') + '_avg'   

    # validation_metric ganz nach vorne
    mask = result.name == 'validation_metric'
    zeilen_ganzvorne = list(set(result[mask].iloc[0]))    
    #print(zeilen_ganzvorne)


    # Zeilen sortieren
    if size == 'small':
        zeilen_vorne =  ['accuracy','recall','specificity','precision','roc_auc','loss',]
        zeilen_hinten = ['epochs','time/epoch','train_time',]    
        result = result.set_index('name').T
        result = pak.move_cols( result, zeilen_vorne )
        result = pak.move_cols( result, zeilen_ganzvorne )        
        result = pak.move_cols( result, zeilen_hinten, -1 )    
        result = result.T.reset_index()

        # Zeilen löschen
        zeilen_verboten = ['validation_metric']
        zeilen_erlaubt = [z for z in (zeilen_ganzvorne + zeilen_vorne + zeilen_hinten) if not z in zeilen_verboten ]
        mask = result.name.isin( zeilen_erlaubt )
        result = result[mask]
        
    else: # size == 'big'
        result = result.sort_values('name')
        zeilen_vorne =  ['validation_metric',]
        zeilen_hinten = ['experiment_path','output_feature_type','output_feature_name','epochs','train_secs','time/epoch','train_time',]     
        result = result.set_index('name').T
        result = pak.move_cols( result, zeilen_vorne )         
        result = pak.move_cols( result, zeilen_hinten, -1 )    
        result = result.T.reset_index()        

    result = pak.reset_index(result)
    result.columns = [c if isinstance(c,str) else 'model_' + str(c) for c in result.columns ]
    return result



def list_datasets():

    datasets = [
#                                                       
#        dataset_name                    input_field_types   output_field_type    
#                                    size                                  note
        ['adult_census_income',      4,  'category, number', 'binary',     'many category fields'], 
        ['agnews',                   5,  'text',             'category',   ''   ],
        ['amazon_review_polarity',   6,  'text',             'binary',     '',  ],
        ['amazon_reviews',           6,  'text',             'category',   '',  ],
        ['dbpedia',                  5,  'text',             'category',   '',  ],
        ['electricity',              4,  'number',           'number',     '',  ],
        ['ethos_binary',             2,  'text',             'binary',     '',  ],
        ['flickr8k',                 3,  'image',            'text',       '',  ],
        ['goemotions',               4,  'text',             'category',   '',  ],
        ['irony',                    3,  'text',             'binary',     '',  ],
        ['mnist',                    4,  'image',            'category',   'Hello world',  ],
        ['mushroom_edibility',       3,  'category',         'binary',     '',  ],
        ['poker_hand',               6,  'category',         'category',   '',  ],
        ['sarcos',                   4,  'number',           'number',     '',  ],
        ['sst2',                     3,  'text',             'binary',     '',  ],
        ['sst3',                     4,  'text',             'category',   '',  ],        
        ['sst4',                     4,  'text',             'category',   '',  ],       
        ['yahoo_answers',            6,  'text',             'category',   '',  ],
        ['yelp_review_polarity',     5,  'text',             'binary',     '',  ],
        ['yelp_reviews',             5,  'text',             'category',   '',  ],
        ['yosemite',                 4,  'time',             'number',     '',  ],
        ['ae_price_prediction',      4,  'category',         'number',     '',  ],     
        ['bookprice_prediction',     3,  'text,category',    'number',     '',  ],    


        
        ]
    result = pak.dataframe(datasets)
    result.columns = ['dataset_name','size','input_field_types','output_field_type','note',]
    return result



def load_dataset(dataset_name, verbose=True):
    '''
    Loads a dataset from Ludwig's dataset zoo. 
    * dataset_name: Name of the dataset. Valid names are listed by list_datasets().
    Returns a DataFrame with training data and a ludwig.dataset_loader.
    '''
    module         = importlib.import_module('ludwig.datasets')
    dataset_loader = getattr(module, dataset_name)
    data_df       = dataset_loader.load(split=False)
        
    #katalog = list_datasets()
    #mask = katalog.dataset_name == dataset_name
    #katalogeintrag = katalog[mask].iloc[0]

    output_feature_names = [col['name'] for col in dataset_loader.config.output_features]   
    if dataset_name == 'agnews':
        output_feature_names = ['class']
    data_df = pak.move_cols(data_df, output_feature_names)     
    data_df = pak.move_cols(data_df, 'split',-1) 
    data_df = pak.drop_cols(data_df,['Unnamed: 0','Unnamed: 1'])
    data_df = pak.change_datatype(data_df)

    if verbose:
        print()
        print(dataset_loader.description())    
        print('output_features:', dataset_loader.config.output_features)  
        
    return data_df, dataset_loader



def analyse_cols(data_df, dataset_loader=None, output_features_size=1):
    '''
    Analyses training data. 
    The information whether a column is input_feature or output_feature comes either from
    * dataset_loader (provided by load_dataset), or from
    * output_features_size (then it is the first n columns of the DataFrame).
    If a dataset_loader is provided, output_features_type is taken from config. Otherwise it's guessed.
    '''

    # analyse_cols
    spalten = ['col_name', 'datatype_short', 'datatype_identified','mem_usage','nunique', 'ndups', 'vmin', 'vmax','n']
    
    analyse = pak.analyse_cols(  pak.sample(data_df, 10000), human_readable=False  )[spalten].iloc[1:]
    mask = analyse.col_name == 'split'
    analyse = pak.drop_rows(analyse,mask)
    analyse['is_output_feature'] = False
    analyse['feature_type'] = ''
    analyse = pak.move_cols(analyse,['is_output_feature','feature_type'],'col_name')
    analyse = pak.reset_index(analyse)

    # is_output_feature
    if dataset_loader is not None:
        output_feature_names = [col['name'] for col in dataset_loader.config.output_features]    
    else:
        output_feature_names = list(data_df.columns)[:output_features_size]
    mask = analyse.col_name.isin(output_feature_names)
    analyse.loc[mask,'is_output_feature'] = True

    # binary feature_type
    mask = analyse['nunique'] == 2
    analyse.loc[mask,'feature_type'] = 'binary'

    # category from number feature_type
    mask0 = analyse['datatype_identified'].isin(['int','float'])    
    mask1 = analyse['nunique'] / analyse['ndups'] < 0.05
    mask2 = analyse['nunique'] <= 30    
    mask3 = analyse.feature_type == ''
    mask = mask0  &  mask1  &  mask2  &  mask3
    analyse.loc[mask,'feature_type'] = 'category'

    # category from text feature_type
    mask0 = analyse['datatype_identified'].isin(['string']) 
    mask1 = analyse['nunique'] / analyse['ndups'] < 0.1
    mask2 = analyse['nunique'] <= 50    
    mask3 = analyse.feature_type == ''
    mask = mask0  &  mask1  &  mask2  &  mask3
    analyse.loc[mask,'feature_type'] = 'category'    

    # numeric feature_type
    mask1 = analyse['datatype_identified'].isin(['int','float'])
    mask2 = analyse.feature_type == ''
    mask = mask1 & mask2
    analyse.loc[mask,'feature_type'] = 'number'

    # text feature_type
    mask1 = analyse['datatype_identified'].isin(['string'])
    mask2 = analyse.feature_type == ''
    mask = mask1 & mask2
    analyse.loc[mask,'feature_type'] = 'text'

    
    # image feature_type  
    try:
        mask1 = analyse['feature_type'] == 'text'
        mask2 = analyse['vmin'].str.endswith(('jpg','png'))
        mask = mask1 & mask2
        analyse.loc[mask,'feature_type'] = 'image'
    except:
        pass

    # date feature_type  
    mask1 = analyse['datatype_identified'].isin(['datetime'])
    mask2 = analyse.feature_type == ''
    mask = mask1 & mask2
    analyse.loc[mask,'feature_type'] = 'date'

    # sequence feature_type  
    mask1 = analyse['datatype_identified'].isin(['list'])
    mask2 = analyse.feature_type == ''
    mask = mask1 & mask2
    analyse.loc[mask,'feature_type'] = 'sequence'    

    # feature_type_guess
    #analyse['feature_type_guess'] = analyse.feature_type.copy()
    #analyse = pak.move_cols(analyse, 'feature_type_guess', 'feature_type')
    
    # überschreibe feature_type_real
    if dataset_loader is not None:
        output_features_real = pak.dataframe(dataset_loader.config.output_features)
        analyse = pak.update_col(analyse, output_features_real, left_on='col_name', right_on='name', col='type', col_rename='feature_type')    
    
    return analyse 





def config0(data_df, dataset_loader=None, output_features_size=1, use_yaml=True):
    '''
    Creates a default config for a given DataFrame. To identify the output_features, use
    * dataset_loader or 
    * output_features_size as in analyse_cols().
    '''
    analyse = analyse_cols(data_df, dataset_loader, output_features_size)
    mask = analyse.is_output_feature
    output_features_size = analyse[mask].shape[0]
    
    # input_features
    spalten = ['col_name','feature_type']
    input_features = analyse[output_features_size:][spalten]
    input_features.columns = ['name','type']
    input_features = input_features.to_dict(orient='records')  

    # output_features
    output_features = analyse[:output_features_size][spalten]
    output_features.columns = ['name','type']
    output_features = output_features.to_dict(orient='records')

    # config
    config = {'input_features':  input_features,
              'output_features': output_features }

    if use_yaml:
        return yaml.dump(config)
    return config



def config1(dataset_loader, use_yaml=True):
    '''
    Wrapper for default_model_config from dataset_loader
    '''
    result = dataset_loader.default_model_config

    if use_yaml:
        result = yaml.dump(result)
        result = result.replace( 'null\n...\n', '' )
        
    return result    



def get_datasets(remove_failed=True):
    '''
    List all datasets available in Ludwig.
    '''
    library_location = pkg_resources.get_distribution('ludwig').location
    directory_within_library = 'ludwig/datasets/configs'
    directory_path = os.path.join(library_location, directory_within_library)
    file_list = os.listdir(directory_path)
    
    elements_to_remove = ['__pycache__','__init__.py']
    result = [x.replace('.yaml','') for x in file_list if x not in elements_to_remove]
    result = sorted(result)

    if not remove_failed:
        return result

    failed =    ['allstate_claims_severity',
                 'amazon_employee_access_challenge',
                 'ames_housing',
                 'bbcnews',
                 'bnp_claims_management',
                 'connect4',
                 'creditcard_fraud',
                 'customer_churn_prediction',
                 'higgs',
                 'ieee_fraud',
                 'imbalanced_insurance',
                 'imdb',
                 'insurance_lite',
                 'jigsaw_unintended_bias100k',
                 'mercedes_benz_greener',
                 'noshow_appointments',
                 'numerai28pt6',
                 'ohsumed_7400',
                 'otto_group_product',
                 'porto_seguro_safe_driver',
                 'reuters_r8',
                 'rossman_store_sales',
                 'santander_customer_satisfaction',
                 'santander_customer_transaction',
                 'santander_value_prediction',
                 'sarcastic_headlines',
                 'synthetic_fraud',
                 'talkingdata_adtrack_fraud',
                 'telco_customer_churn',
                 'temperature',
                 'titanic',
                 'twitter_bots',
                 'walmart_recruiting',
                 'wmt15',
                 'fever',
                 'forest_cover',
                 'google_quest_qa',
                 'imdb_genre_prediction',
                 'kdd_appetency',
                 'kdd_churn',
                 'kdd_upselling']    

    result = [x for x in result if x not in failed]
    return result





def scan_datasets(dataset_names, get_elements_to_remove=False):
    '''
    Scan and analyse Ludwig's database zoo. 
    '''
    result = []
    for dataset in dataset_names:
        
        # load_dataset
        print('loading',dataset)
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")            
                data_df, dataset_loader = load_dataset(dataset, verbose=False)  
        except:
            elements_to_remove += [dataset]

        # analysieren 
        analyse = analyse_cols(data_df, dataset_loader) 
        mask = analyse.is_output_feature
        output_features_size = analyse[mask].shape[0]        

        # output_features and input_features 
        spalten = ['col_name','feature_type','mem_usage']
        output_features = analyse[:output_features_size][spalten]           
        input_features = analyse[output_features_size:][spalten]        

        # Ausgabewerte berechnen
        input_features_type  = ', '.join(pak.group_and_agg( input_features,  ['feature_type','mem_usage'], ['','sum']).sort_values('mem_usage_sum', ascending=False).feature_type)
        output_features_type = ', '.join(pak.group_and_agg( output_features, ['feature_type','mem_usage'], ['','sum']).sort_values('mem_usage_sum', ascending=False).feature_type)        
        config1 = yaml.dump(dataset_loader.default_model_config)
        config1 = config1.replace('null\n...\n','')

        # Ausgabe
        result += [{'dataset_name': dataset, 
                    'rows': bpy.human_readable_number(data_df.shape[0]) ,
                    'in_size'  : input_features.shape[0],                        
                    'out_size' : output_features.shape[0],

                    'in_name'  : ', '.join(input_features.col_name),                    
                    'out_name' : ', '.join(output_features.col_name),                    

                    'in_type'  : input_features_type,  # ', '.join(input_features.feature_type),                      
                    'out_type' : output_features_type, # ', '.join(output_features.feature_type),                    
                                 
                    'description': dataset_loader.description(),
                    'config1': config1(dataset_loader),

                   }]
    

    result = pak.dataframe(result) 

    if get_elements_to_remove:
        return result, elements_to_remove
    return result















import logging, os, shutil, time, yaml, warnings, json
import pandas as pd
import bpyth as bpy

from munch import DefaultMunch

from ludwig.api       import LudwigModel
from ludwig.visualize import learning_curves
from ludwig.visualize import compare_performance
from ludwig.visualize import confusion_matrix
from ludwig.visualize import roc_curves_from_test_statistics
from torchinfo        import summary

from bludwig.helper import *

# logging.DEBUG   == 10
# logging.INFO    == 20
# logging.WARNING == 30


# LudwigJob wird einmal instanziiert
#
class LudwigJob():

    
#####################################################################################################
# Basics
#     
    
    def __init__( self, configs=[], experiment_name=None, verbose=False ):  
        '''
        LudwigJob wird einmal instanziiert, z.B. so:
        ludwig_job = bludwig.LudwigJob( configs=configs, verbose=True) 
        * configs: list of Ludwig-configs as YAML-String, Path to yaml-file or yaml-object
            
        '''
        # Parameter         
        if experiment_name is None:
            experiment_name = 'ex'
        self.experiment_name = experiment_name
        self.verbose = verbose        

        # configs ggf. in YAML wandeln
        self.configs = []
        for config in configs:
            if isinstance(config, str):
                if config.count('\n') > 2: 
                    self.configs += [yaml.safe_load(config)]
                else:
                    self.configs += [config]   

        try:
            import google.colab
            self.in_colab = True 
        except:
            self.in_colab = False     

        # Tensorflow warnings unterdrücken
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'    

        # gpu_info
        if verbose:
            gpu_info()

        # aktuelles model
        self.model = None
        self.model_no = None
        self.cuda = None
        self.output_feature_name = ''

        # job
        self.train_jobs = []
        self.model_names = []
        self.model_paths = []

        # train_log
        self.train_log_raw = pd.DataFrame()

        # results
        self.train_stats      = []
        self.test_stats       = []
        self.output_dirs      = []       

        if len(configs) > 0:
            print()
            print('{} configs loaded'.format(len(configs)))



    def __str__(self):
        result = f'''LudwigJob object
        experiment_name:     {self.experiment_name}
        output_feature_name: {self.output_feature_name}
        train_jobs:          {self.train_jobs}
        model_names:         {self.model_names}
        model_paths:         {self.model_paths}        
        output_dirs:         {self.output_dirs}
        model_no:            {self.model_no}   
        cuda:                {self.cuda}             
        '''
        return result


    
    def load_from_results(self):    
        '''
        Loads data from results directory
        '''
        results_dir = os.listdir('results')
        self.output_dirs = [ 'results/' + d                for d in results_dir ]
        self.train_jobs  = [ int(d.split('_')[1])          for d in results_dir ]
        self.model_names = [ 'model_' + str(j)             for j in self.train_jobs ]    
        self.model_paths = [d + '/model'                   for d in self.output_dirs]
        
        train_stats = [ d + '/training_statistics.json'    for d in self.output_dirs ]
        train_stats = [ json.load( open(p) )               for p in train_stats]
        self.train_stats = [ DefaultMunch.fromDict(d)      for d in train_stats]

        test_stats = [ d + '/test_statistics.json'         for d in self.output_dirs ]
        test_stats = [ json.load( open(p) )                for p in test_stats]
        self.test_stats = [ DefaultMunch.fromDict(d)       for d in test_stats]

        # output_feature_name
        a = list(self.test_stats[0].keys())
        a.remove('combined')
        self.output_feature_name = a[0]   
        # nochmal gegenprüfen
        b = list(self.train_stats[0]['test'].keys())
        b.remove('combined')
        assert self.output_feature_name in b


    
    def load_model(self, model_no, cuda=True):
        '''
        Loads a model identified by model number.
        * cuda: Use cuda, if available
        '''
        if model_no == self.model_no and cuda == self.cuda:
            return
            
        self.model = LudwigModel.load( self.model_paths[model_no] )
        if cuda and torch.cuda.is_available():
            self.model.model.to('cuda')
            self.cuda = True
        else:
            self.model.model.to('cpu')   
            self.cuda = False
        self.model_no = model_no




        
#####################################################################################################
# print_models
#

    def print_model(self, model_no=None):
        '''
        Uses torchinfo.summary to print a model
        '''
        if model_no is None:
            model_no = self.model_no
        self.load_model(model_no, cuda=False)
        print( '### {} ###'.format(self.model_names[model_no]))
        print( summary(self.model.model, 
                       input_data=[self.model.model.get_model_inputs()], 
                       depth=20, 
                       col_names=['input_size','output_size','num_params','trainable'] 
                      ) )
        print('\n'*3)   


    
    def print_models(self):        
        for model_no in self.train_jobs:
            self.print_model(model_no)

            
        
#####################################################################################################
# experiment
#
    
    def experiment(self, train_jobs, dataset):
        '''
        train and evaluate a list of Ludwig models
        '''
        self.train_jobs = train_jobs
        self.model_names = ['model_' + str(c) for c in train_jobs]
        
        
        for config_no in train_jobs:
            experiment_subname = self.experiment_name + '_' + str(config_no) 
            experiment_path = 'results/' + experiment_subname + '_run'  
            print()
            print( 'Training config_no {} >> {}'.format( config_no, experiment_path) )   
            
            # Zielverzeichnis rekursiv löschen
            try:
                shutil.rmtree(experiment_path)
            except:
                pass

            logging_level = 20 if self.verbose else 30

            # lade model
            self.model               = LudwigModel(config=self.configs[config_no], logging_level=logging_level)   
            self.model_no            = config_no
            self.cuda                = torch.cuda.is_available()
            self.output_feature_name = self.model.config['output_features'][0]['name']
            self.output_feature_type = self.model.config['output_features'][0]['type']

            # trainiere
            start_time = time.time()     
            test_stat, train_stat, _, output_dir = self.model.experiment( dataset=dataset, experiment_name=experiment_subname)
            self.test_stats  += [test_stat]
            self.train_stats += [train_stat]
            self.output_dirs += [output_dir]
            self.model_paths += [output_dir + '/model']

            # logge config
            train_secs        = round( time.time() - start_time )
            train_time        = bpy.human_readable_seconds( train_secs )
            epochs            = len(train_stat.test['combined']['loss']) 
            validation_metric = self.model.config['trainer']['validation_metric']
            log = pak.dataframe([
                [ config_no, 'train_secs',          train_secs        ],
                [ config_no, 'train_time',          train_time        ],
                [ config_no, 'epochs',              epochs            ],
                [ config_no, 'time/epoch',          bpy.human_readable_seconds( train_secs/epochs ) ],     
                [ config_no, 'validation_metric',   validation_metric ],     
                [ config_no, 'experiment_path',     experiment_path   ], 
                [ config_no, 'output_feature_name', self.output_feature_name   ],    
                [ config_no, 'output_feature_type', self.output_feature_type   ],                    
                
            ])
            log.columns = ['config_no','name','value']
            self.train_log_raw = pak.add_rows( self.train_log_raw, log ) 
            print('train_time:',train_time) 
            print()

        # logge Gesamtprozess
        self.train_log_raw = pak.add_rows( self.train_log_raw, entwirre_test_stat(self.test_stats))            



        
#####################################################################################################
# predict
#

    def predict(self, data, merge=True):
        pred, _ = self.model.predict(data)
        if not merge:
            return pred
            
        pred.index = data.index
        result = pd.merge( data, pred, left_index=True, right_index=True)
        result = pak.move_cols( result, [self.output_feature_name, self.output_feature_name + '_predictions'])
        result.rename(columns=lambda x: x.replace('_predictions', '_pred').replace('probabilities', 'prob').replace('probability', 'prob'), inplace=True)
        return result        


        
#####################################################################################################
# train_log
#

    def train_log(self):
        ''' returns small log'''
        if self.train_log_raw.shape[0] > 0:
            result = prepare_train_log(self.train_log_raw, size='small')
            return result
        else:
            zeilen = ['roc_auc','accuracy','recall','specificity','precision','loss','epochs','time/epoch','train_time']
            result = pd.DataFrame(zeilen)
            result.columns = ['name']
            return result

    
    
    def train_log_big(self):
        ''' returns bigger log'''
        return prepare_train_log(self.train_log_raw, size='big')        


        
    def train_log_to_csv(self):
        '''
        Saves train_log_big to csv file
        Shows train_log (small version)
        '''
        t = self.train_log_big()
        if self.in_colab:
            t.to_csv('train_log_colab.csv', index=False)
        else:
            if torch.cuda.is_available():
                t.to_csv('train_log_GPU.csv', index=False)    
            else:
                t.to_csv('train_log_CPU.csv', index=False)
                
        return self.train_log()


        
        
#####################################################################################################
# Visualisierungen
#

    def compare_performance(self, output_feature_name=None):

        # Kein bestimmer output_feature_name angefragt >> Default
        if output_feature_name is None:
            output_feature_name = self.output_feature_name
        
        # test_stats_small (verhindert Fehler)
        test_stats_small = []
        keys_to_keep = list(self.train_log().name)
        for stat in self.test_stats:
            r = {key: value  for key, value in stat[output_feature_name].items()  if key in keys_to_keep}
            r = { output_feature_name: r }    
            test_stats_small += [r]        

        # ausgeben
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            compare_performance( test_stats_small, model_names=self.model_names, output_feature_name=output_feature_name )


    
    def learning_curves(self, output_feature_name=None):
        
        # Kein bestimmer output_feature_name angefragt >> Default
        if output_feature_name is None:
            output_feature_name = self.output_feature_name    
            
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            learning_curves( self.train_stats, model_names=self.model_names, output_feature_name=output_feature_name)     


    
    def confusion_matrix(self, output_feature_name=None, normalize=True, top_n_classes=[10]):

        # Kein bestimmer output_feature_name angefragt >> Default
        if output_feature_name is None:
            output_feature_name = self.output_feature_name  

        #load_model
        if self.model is None:
            try:
                self.load_model(0)
            except:
                print('No model loaded')
            
        # confusion_matrix
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            confusion_matrix( self.test_stats, 
                              self.model.training_set_metadata, 
                              output_feature_name=output_feature_name, 
                              top_n_classes=top_n_classes, 
                              model_names=self.model_names, 
                              normalize=True )


    
    def roc_curves(self, output_feature_name=None):
        '''
    
        '''
        # Kein bestimmer output_feature_name angefragt >> Default
        if output_feature_name is None:
            output_feature_name = self.output_feature_name  
            
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                roc_curves_from_test_statistics(self.test_stats, output_feature_name=output_feature_name)   
            except KeyError:
                print('No roc_curve found')
                pass
            except TypeError:
                print('No roc_curve found')
                pass                
            except:
                raise  # Löst die Ausnahme erneut aus        
                



